#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <unistd.h>
#include <errno.h>
#include <vgagames3.h>
#include "main.h"
#include "obj-mine.h"

#define MAX_HEALTH 3

void getofc_mine(struct vg3_ofunc_objfunc *);

static struct vg3_ofunc_object * f_new(void *, unsigned int, ...);
static void f_free(void *, struct vg3_ofunc_object *);
static void f_run(void *, struct vg3_ofunc_object *);
static void f_draw(void *, struct vg3_ofunc_object *);


/* +++ get-function +++ */

void
getofc_mine(struct vg3_ofunc_objfunc *ofc)
{
  if (ofc == NULL) { return; }

  snprintf(ofc->oid, sizeof(ofc->oid), "%s", get_oid_name(OID_NAME_MINE));
  ofc->f_new = f_new;
  ofc->f_free = f_free;
  ofc->f_run = f_run;
  ofc->f_draw = f_draw;
}


/* +++ object-instance-functions +++ */

/* new-function
 * variable parameters:
 * - int: x-position (centered)
 * - int: y-position (centered)
 * - int: x-moving
 * - int: y-moving
 * - int: rotating-value
 */
static struct vg3_ofunc_object *
f_new(void *vmain, unsigned int iparent, ...)
{
  struct g_main *gmain = vmain;
  struct vg3_ofunc_object *objp;
  struct g_obj_mine *gobj;
  struct vg3_coll coll;
  va_list ap;
  int xm, ym, xd, yd, rot;

  if (gmain == NULL) { VG3_seterror(EINVAL, strerror(EINVAL)); return NULL; }

  /* get arguments */
  va_start(ap, iparent);
  xm = va_arg(ap, int);
  ym = va_arg(ap, int);
  xd = va_arg(ap, int);
  yd = va_arg(ap, int);
  rot = va_arg(ap, int);
  va_end(ap);

  /* create private struct for object */
  gobj = calloc(1, sizeof(*gobj));
  if (gobj == NULL) { VG3_seterror(ENOMEM, strerror(errno)); return NULL; }

  /* load sprites */
  gobj->images.sprt = VG3_sprite_load(gmain->wstruct, FILES_DIR "/bmp/mine/mine.sprite");
  if (gobj->images.sprt == NULL) { return NULL; }
  gobj->images.hit = VG3_sprite_load(gmain->wstruct, FILES_DIR "/bmp/hit/hit.sprite");
  if (gobj->images.hit == NULL) { return NULL; }
  gobj->images.expl = VG3_sprite_load(gmain->wstruct, FILES_DIR "/bmp/expl/expl2.sprite");
  if (gobj->images.expl == NULL) { return NULL; }

  gobj->pingsnd = VG3_audio_load(gmain->wstruct, FILES_DIR "/sound/ping.wav", 100, VGAG3_AUDIO_VOLUME_SOUND);
  if (gobj->pingsnd == 0) { return NULL; }

  VGAG3_IMAGE_ATTRIBUTES_DEFAULT(&gobj->iattr);
  gobj->rot0 = gobj->rot1 = rot;

  /* set position */
  VG3_sprite_imagesize(gobj->images.sprt, "avg", 100, &gobj->rect.w, &gobj->rect.h);
  gobj->rect.x = xm - gobj->rect.w / 2;
  gobj->rect.y = ym - gobj->rect.h / 2;
  /* set collision rectangle with 80% of width and height of image-rectangle */
  gobj->crect = gobj->rect;
  gobj->crect = VG3_correct_imageposition(&gobj->crect, gobj->rect.w * 80 / 100, gobj->rect.h * 80 / 100);
  /* set collision rectangle for long-distance-quadtree with 500% of width and height of image-rectangle */
  gobj->crect_long = gobj->rect;
  gobj->crect_long = VG3_correct_imageposition(&gobj->crect_long, gobj->rect.w * 500 / 100, gobj->rect.h * 500 / 100);

  /* set speed */
  gobj->xdelta = -10 + xd;
  gobj->ydelta = 0 + yd;

  /* set health */
  gobj->health = MAX_HEALTH;

  /* create and fill mine-instance */
  objp = calloc(1, sizeof(*objp));
  if (objp == NULL) { VG3_seterror(ENOMEM, strerror(errno)); return NULL; }
  snprintf(objp->oid, sizeof(objp->oid), "%s", get_oid_name(OID_NAME_MINE));
  objp->subid = 0;
  objp->drawlevel = 2;
  objp->instanceid = 0;  /* will be set in VG3_ofunc_objlist_insert() */
  objp->ostruct = gobj;

  /* insert mine-instance into list of object-instances */
  VG3_ofunc_objlist_insert(gmain->ofstruct, objp);

  /* insert mine-instance (collision-rectangle) into quadtrees */
  memset(&coll, 0, sizeof(coll));
  coll.rect = gobj->crect;
  snprintf(coll.oid, sizeof(coll.oid), "%s", objp->oid);
  coll.optr = objp;
  VG3_coll_q_insert(gmain->qdtr, &coll);
  coll.rect = gobj->crect_long;
  VG3_coll_q_insert(gmain->game.qdtr_long, &coll);

  return objp;
}


/* free-function */
static void
f_free(void *vmain, struct vg3_ofunc_object *objp)
{
  struct g_main *gmain = vmain;
  struct g_obj_mine *gobj;

  if (gmain == NULL || objp == NULL) { return; }

  gobj = (struct g_obj_mine *)objp->ostruct;

  /* remove mine-instance from quadtrees */
  VG3_coll_q_remove(gmain->qdtr, objp);
  VG3_coll_q_remove(gmain->game.qdtr_long, objp);

  /* remove mine-instance from list of object-instances */
  VG3_ofunc_objlist_remove(gmain->ofstruct, objp);

  /* inform object-management if destroyed */
  if (gobj->health <= 0) {
    struct fdata_number fdnumber;
    memset(&fdnumber, 0, sizeof(fdnumber));
    fdnumber.flag = FDATA_HIT;
    VG3_ofunc_mgmt_data(gmain->ofstruct, gmain, get_oid_name(OID_NAME_MINE), &fdnumber);
  }

  /* free mine-instance */
  if (gobj->images.sprt != NULL) { VG3_sprite_free(gobj->images.sprt); }
  if (gobj->images.hit != NULL) { VG3_sprite_free(gobj->images.hit); }
  if (gobj->images.expl != NULL) { VG3_sprite_free(gobj->images.expl); }
  if (gobj->pingsnd > 0) { VG3_audio_unload(gmain->wstruct, gobj->pingsnd); }
  free(gobj);
  free(objp);
}


/* run-function */
static void
f_run(void *vmain, struct vg3_ofunc_object *objp)
{
  const int moving_factor = 100;
  struct g_main *gmain = vmain;
  struct g_obj_mine *gobj;
  struct vg3_coll coll;
  int erg;

  if (gmain == NULL || objp == NULL) { return; }

  gobj = (struct g_obj_mine *)objp->ostruct;

  /* remove mine-instance from quadtrees */
  VG3_coll_q_remove(gmain->qdtr, objp);
  VG3_coll_q_remove(gmain->game.qdtr_long, objp);

  /* check if mine is exploding */
  if (gobj->health <= 0) { return; }

  /* check if mine is gone */
  if (gobj->rect.x < -gobj->rect.w) { f_free(gmain, objp); return; }
  if (gobj->rect.y < -gobj->rect.h) { f_free(gmain, objp); return; }
  if (gobj->rect.y > gmain->winh) { f_free(gmain, objp); return; }

  /* move mine-instance and check for collisions with gobj->crect */
  erg = VG3_move_object_check_collision(gmain, gmain->ofstruct, gmain->qdtr, objp, &gobj->crect, moving_factor, &gobj->xdelta, &gobj->ydelta, &gobj->xremainder, &gobj->yremainder);
  if (erg < 0) {  /* error: destroy mine-instance */
    fprintf(stderr, "moving mine: %s\n", VG3_error());
    f_free(gmain, objp);
    return;
  }
  if (erg == VGAG3_COLL_RETURN_DEAD) { return; }  /* instance has been freed */
  if (erg == VGAG3_COLL_RETURN_CATCHED) { f_free(gmain, objp); return; }

  /* x/y-value of gobj->crect could be modified, correct it also in gobj->rect */
  gobj->rect = VG3_correct_imageposition(&gobj->crect, gobj->rect.w, gobj->rect.h);
  /* and correct it in the long-distance-quadtree */
  gobj->crect_long = VG3_correct_imageposition(&gobj->crect, gobj->crect_long.w, gobj->crect_long.h);

  /* check long-distance-quadtree */
  if (check_long_distance(gmain, objp, &gobj->crect_long) < 0) {
    fprintf(stderr, "mine: check long-distance-quadtree: %s\n", VG3_error());
    f_free(gmain, objp);
    return;
  }

  /* insert mine-instance (collision-rectangle) into quadtrees again */
  memset(&coll, 0, sizeof(coll));
  coll.rect = gobj->crect;
  snprintf(coll.oid, sizeof(coll.oid), "%s", objp->oid);
  coll.optr = objp;
  VG3_coll_q_insert(gmain->qdtr, &coll);
  coll.rect = gobj->crect_long;
  VG3_coll_q_insert(gmain->game.qdtr_long, &coll);
}


/* draw-function */
static void
f_draw(void *vmain, struct vg3_ofunc_object *objp)
{
  struct g_main *gmain = vmain;
  struct g_obj_mine *gobj;
  struct vg3_image *img;
  struct vg3_image_attributes iattr;

  if (gmain == NULL || objp == NULL) { return; }

  gobj = (struct g_obj_mine *)objp->ostruct;

  /* draw mine-instance */
  if (gobj->health <= 0) {  /* exploding */
    if (!VG3_sprite_get(gobj->images.expl, &img, &iattr, NULL)) { f_free(gmain, objp); return; }
  } else {
    struct vg3_image_attributes mattr;
    if (!VG3_sprite_get(gobj->images.sprt, &img, &mattr, NULL)) {
      VG3_sprite_rewind(gobj->images.sprt);
      VG3_sprite_get(gobj->images.sprt, &img, &mattr, NULL);
    }
    gobj->iattr.rotate += gobj->rot1;
    gobj->iattr.rotate %= 360;
    VG3_image_attr_sum(&iattr, &mattr, &gobj->iattr);
  }
  if (img != NULL) {
    VG3_image_copy(gmain->wstruct, NULL, img, gobj->rect.x + gobj->rect.w / 2, gobj->rect.y + gobj->rect.h / 2, &iattr, 0);
    if (gobj->health > 0 && gobj->hit_y > 0) {  /* being hit */
      if (!VG3_sprite_get(gobj->images.hit, &img, &iattr, NULL)) {
        VG3_sprite_rewind(gobj->images.hit);
        gobj->hit_y = 0;
      } else if (img != NULL) {
        VG3_image_copy(gmain->wstruct, NULL, img, gobj->rect.x + gobj->rect.w / 2, gobj->rect.y + gobj->hit_y, &iattr, 0);
        draw_health(vmain, MAX_HEALTH, gobj->health, 0);
      }
    }
  }

  gobj->rot1 = gobj->rot0;
}
